# Copyright 2018 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from os.path import join, dirname, expanduser, exists, isfile
from time import sleep
import ovos_config.meta as _ovos_config
from ovos_utils.xdg_utils import xdg_config_dirs, xdg_config_home, xdg_data_dirs, xdg_data_home, xdg_cache_home


def get_xdg_config_dirs(folder=None):
    """ return list of possible XDG config dirs taking into account ovos.conf """
    folder = folder or _ovos_config.get_xdg_base()
    xdg_dirs = xdg_config_dirs() + [xdg_config_home()]
    return [join(path, folder) for path in xdg_dirs]


def get_xdg_data_dirs(folder=None):
    """ return list of possible XDG data dirs taking into account ovos.conf """
    folder = folder or _ovos_config.get_xdg_base()
    return [join(path, folder) for path in xdg_data_dirs()]


def get_xdg_config_save_path(folder=None):
    """ return base XDG config save path taking into account ovos.conf """
    folder = folder or _ovos_config.get_xdg_base()
    return join(xdg_config_home(), folder)


def get_xdg_data_save_path(folder=None):
    """ return base XDG data save path taking into account ovos.conf """
    folder = folder or _ovos_config.get_xdg_base()
    return join(xdg_data_home(), folder)


def get_xdg_cache_save_path(folder=None):
    """ return base XDG cache save path taking into account ovos.conf """
    folder = folder or _ovos_config.get_xdg_base()
    return join(xdg_cache_home(), folder)


def find_user_config():
    """ return user config full file path taking into account ovos.conf """
    path = join(get_xdg_config_save_path(), _ovos_config.get_config_filename())
    if isfile(path):
        return path
    old, path = get_config_locations(default=False, web_cache=False,
                                     system=False, old_user=True,
                                     user=True)
    if isfile(path):
        return path
    if isfile(old):
        return old
    return path


def get_config_locations(default=True, web_cache=True, system=True,
                         old_user=True, user=True):
    """return list of all possible config files paths sorted by priority taking into account ovos.conf"""
    locs = []
    ovos_cfg = _ovos_config.get_ovos_config()
    if default:
        locs.append(ovos_cfg["default_config_path"])
    if system:
        locs.append(f"/etc/{ovos_cfg['base_folder']}/{ovos_cfg['config_filename']}")
    if web_cache:
        locs.append(get_webcache_location())
    if old_user:
        locs.append(f"~/.{ovos_cfg['base_folder']}/{ovos_cfg['config_filename']}")
    if user:
        locs.append(f"{get_xdg_config_save_path()}/{ovos_cfg['config_filename']}")
    return locs


def get_webcache_location():
    """ return remote config cache full file path taking into account ovos.conf """
    return join(get_xdg_config_save_path(), 'web_cache.json')


def get_xdg_config_locations():
    """ return list of possible XDG config full file paths taking into account ovos.conf """
    # This includes both the user config and
    # /etc/xdg/mycroft/mycroft.conf
    xdg_paths = list(reversed(
        [join(p, _ovos_config.get_config_filename())
         for p in get_xdg_config_dirs()]
    ))
    return xdg_paths


def find_default_config():
    """ find where mycroft is installed and return the path to the default mycroft.conf
    if mycroft is not found then return the bundled file in ovos_config package"""
    from ovos_utils.system import search_mycroft_core_location
    try:
        mycroft_root = search_mycroft_core_location()
        if not mycroft_root:
            raise FileNotFoundError("Couldn't find mycroft core root folder.")
        cfg = join(mycroft_root, "mycroft", "configuration", "mycroft.conf")
        if isfile(cfg):
            return cfg
    except FileNotFoundError:
        pass

    # mycroft-core not found
    return join(dirname(__file__), "mycroft.conf")


DEFAULT_CONFIG = _ovos_config.get_ovos_config()['default_config_path']
SYSTEM_CONFIG = os.environ.get('MYCROFT_SYSTEM_CONFIG',
                               f'/etc/{_ovos_config.get_xdg_base()}/'
                               f'{_ovos_config.get_config_filename()}')
# TODO: remove in 22.02
# Make sure we support the old location still
# Deprecated and will be removed eventually
OLD_USER_CONFIG = join(expanduser('~'), '.' + _ovos_config.get_xdg_base(),
                       _ovos_config.get_config_filename())
USER_CONFIG = join(get_xdg_config_save_path(),
                   _ovos_config.get_config_filename())
REMOTE_CONFIG = "mycroft.ai"
WEB_CONFIG_CACHE = os.environ.get('MYCROFT_WEB_CACHE') or \
                   get_webcache_location()


def __ensure_folder_exists(path):
    """ Make sure the directory for the specified path exists.

        Args:
            path (str): path to config file
     """
    directory = dirname(path)
    if not exists(directory):
        try:
            os.makedirs(directory)
        except:
            sleep(0.2)
            if not exists(directory):
                try:
                    os.makedirs(directory)
                except Exception as e:
                    pass


__ensure_folder_exists(WEB_CONFIG_CACHE)
__ensure_folder_exists(USER_CONFIG)



from dateutil.tz import gettz, tzlocal
import ovos_config

# lingua_franca is optional and might not be installed
# exceptions should only be raised in the parse and format utils


try:
    import lingua_franca as LF
except ImportError:
    LF = None

_lang = None
_default_tz = None


def get_primary_lang_code(config=None):
    global _lang
    if LF:
        return LF.get_primary_lang_code()
    if not _lang:
        config = config or ovos_config.Configuration()
        _lang = config.get("lang", "en-us")
    return _lang.split("-")[0]


def get_default_lang(config=None):
    """
    Get the default localized language code (i.e. `en-us`)
    @param config: Configuration (default is Configuration())
    @return: lowercase BCP-47 language code
    """
    global _lang
    if LF and LF.get_default_loc():
        return LF.get_default_loc()
    if not _lang:
        config = config or ovos_config.Configuration()
        _lang = config.get("lang", "en-us")
    return _lang


def set_default_lang(lang):
    global _lang
    _lang = lang
    if LF:
        LF.set_default_lang(lang)


def get_config_tz():
    code = ovos_config.Configuration()["location"]["timezone"]["code"]
    return gettz(code)


def get_default_tz():
    # if default was set at runtime use it else use the timezone from .conf
    return _default_tz or get_config_tz()


def set_default_tz(tz=None):
    """ configure LF """
    global _default_tz
    tz = tz or get_config_tz() or tzlocal()
    _default_tz = tz
    if LF:
        # tz added in recently, depends on version
        try:
            LF.time.set_default_tz(tz)
        except:
            pass


def load_languages(langs):
    if LF:
        LF.load_languages(langs)


def load_language(lang):
    if LF:
        LF.load_language(lang)


def setup_locale(lang=None, tz=None):
    lang_code = lang or ovos_config.Configuration().get("lang", "en-us")
    # Load language resources, currently en-us must also be loaded at all times
    load_languages([lang_code, "en-us"])
    # Set the active lang to match the configured one
    set_default_lang(lang_code)
    # Set the default timezone to match the configured one
    set_default_tz(tz)


# mycroft-core backwards compat LF only interface
def set_default_lf_lang(lang_code="en-us"):
    """Set the default language of Lingua Franca for parsing and formatting.

    Note: this is a temporary method until a global set_default_lang() method
    can be implemented that updates all Mycroft systems eg STT and TTS.
    It will be deprecated at the earliest possible point.

    Args:
        lang (str): BCP-47 language code, e.g. "en-us" or "es-mx"
    """
    return set_default_lang(lang_code)



#!/bin/env python3
from typing import Any, Tuple

import rich_click as click
from rich.console import Console
from rich.prompt import Prompt
from rich.table import Table

from ovos_config import Configuration, LocalConf
from ovos_config.locations import USER_CONFIG

CONFIG = Configuration()
CONFIGS = [("Joined", CONFIG),
           ("Sytem", CONFIG.system),
           ("User", LocalConf(USER_CONFIG)),
           ("Remote", CONFIG.remote)]
SECTIONS = [k for k, v in CONFIG.items() if isinstance(v, dict)] + ["base"]


def drawTable(dic: dict, table: Table, level: int = 0) -> None:
    for key, value in dic.items():
        s = f'{key:>{4*level+len(key)}}'
        if not isinstance(value, dict):
            table.add_row(s, str(value))
        else:
            if level == 0:
                table.add_section()
            table.add_row(f'[red]{s}[/red]')
            drawTable(value, table, level+1)
    return


def dictDepth(dic: dict, level: int = 1) -> int:
     
    if not isinstance(dic, dict) or not dic:
        return level
    return max(dictDepth(dic[key], level + 1)
               for key in dic)


def walkDict(dic: dict,
             key: str,
             full_path: bool = False,
             path: Tuple = (),
             found: bool = False):
    for k in dic.keys():
        if key.lower() in k.lower():
            found = True
            if not full_path:
                yield path+(k,), dic[k]
                found = False

        # endpoint
        if type(dic[k]) != dict:
            if found:
                yield path+(k,), dic[k]
        else:
            yield from walkDict(dic[k],
                                key,
                                full_path,
                                path+(k,),
                                found)
        found = False


def pathGet(dic: dict, path: str) -> Any:
    path = path.lstrip("/")
    for item in path.split("/"):
        dic = dic[item]
    return dic


def pathSet(dic: dict, path: str, value: Any) -> None:
    _path = path.lstrip("/").split("/")
    _key = _path.pop(-1)
    for entry in _path:
        if entry not in dic:
            dic[entry] = {}
        dic = dic[entry]
    dic[_key] = value


click.rich_click.STYLE_ARGUMENT = "dark_red"
click.rich_click.STYLE_OPTION = "dark_red"
click.rich_click.STYLE_SWITCH = "indian_red"
click.rich_click.COMMAND_GROUPS = {
    "ovos-config": [
        {
            "name": "Show configuration tables (Joined/User/System/Remote)",
            "commands": ["show"],
            "table_styles": {
                "row_styles": ["white"],
                "padding": (0, 2),
                "title_justify": "left"
            },
        },
        {
            "name": "Get specific key(s)",
            "commands": ["get"],
            "table_styles": {
                "row_styles": ["white"],
                "padding": (0, 2),
            },
        },
        {
            "name": "Setting user values",
            "commands": ["set"],
            "table_styles": {
                "row_styles": ["white"],
                "padding": (0, 3),
            },
        },
    ]
}

console = Console()

@click.group()
def config():
    """\b
    Small helper tool to quickly show, get or set config values

    `ovos-config [COMMAND] --help` for further information about the specific command ARGUMENTS
    \b
    """
    pass


@config.command()
@click.option("--user", "-u", is_flag=True, help="User Configuration")
@click.option("--system", "-s", is_flag=True, help="System Configuration")
@click.option("--remote", "-r", is_flag=True, help="Remote Configuration")
@click.option("--section", default="", show_default=False, help="Choose a specific section from the underlying configuration")
@click.option("--list-sections", "-l", is_flag=True, help="List the sections based on the underlying configuration")
def show(user, system, remote, section, list_sections):
    """\b
    By ommiting a specific configuration a joined configuration table is shown. (which is the one ultimately gets loaded by ovos)
    \b
    Based on this consideration you can further trim the table by section.
    If the sections are unknown you may want to list them.
    \b
    Examples: 
    ovos-config show                                    # shows all the configuration values in a table format
    ovos-config show -s -l                              # shows the sections of the system configuration
    ovos-config show -u --section base                  # shows only the base (ie. top level) values of the user configuration 

    note: joining pattern: user > system > remote > default
    \b
    """
    if not any([user, system, remote]):
        name, config = CONFIGS[0]
    elif system:
        name, config = CONFIGS[1]
    elif user:
        name, config = CONFIGS[2]
    elif remote:
        name, config = CONFIGS[3]
    

    # based on chosen configuration
    if name != "Joined":
        _sections = [k for k, v in config.items() if isinstance(v, dict)]
        if len([k for k, v in config.items() if not isinstance(v, dict)]):
            _sections.append("base")
    else:
        _sections = SECTIONS

    if list_sections:        
        console.print(f"Available sections ({name} config): " + " ".join(_sections))
        exit()

    if section:
        # general info that no such key exists
        if section not in SECTIONS:
            console.print(f"The section `{section}` doesn't exist. Please chose"
                          f" from {' '.join(SECTIONS)}")
            exit()
        # based on chosen configuration
        elif section not in _sections:
            found_in = [f"`{_name}`" for _name, _config in CONFIGS
                        if section in _config and _name != name]            
            console.print(f"The section `{section}` doesn't exist in the {name} "
                          f"Configuration. It is part of the {'/'.join(found_in)} "
                          "Configuration though")
            exit()
        elif section == "base":
            _config = {k: v for k, v in config.items() if not isinstance(v, dict)}
        else:
            _config = config[section]
    else:
        # sorted dict based on depth
        _config = {key: value for key, value in
                    sorted(config.items(), key=lambda item: dictDepth(item[1]))}

    section_info = f", Section: {section}" if section else ""
    additional_info = f"(Configuration: {name}{section_info})"
    
    table = Table(show_header=True, header_style="bold red")
    table.add_column(f"Configuration keys {additional_info}", min_width=60)
    table.add_column("Value", min_width=20)

    drawTable(_config, table)
    console.print(table)    

@config.command()
@click.option("--key", "-k", required=True, help="the key (or parts thereof) which should be searched")
def get(key):
    """\b
    Search for config keys in the (joined) configuration
    \b
    Meant to either loosely search for keys resp. parts thereof or specific dictionary paths (form: `/path/to/key`)
    The loose search will output a list of found keys - if there are multiple - that match the query (full or in part)
    The strict search performs a query to a specific path and will only output the value. (usefull for shell scripting)
    \b
    Examples: 
    ovos-config get -k lang                              # get all lang key values across the configuration
    ovos-config get -k /tts/module                       # get the key at the position specified
    """
    strict = True if "/" in key else False
    if strict:
        if not key.startswith("/"):
            console.print("A strict key search has to start with `/` (root)")
            exit()

        value = pathGet(CONFIG, key)
        # Configuration is excepting ValueError itself
        if value is None:
            exit()

        console.print(f"[red]{value}[/red]")
    else:
        values = list(walkDict(CONFIG, key))
        if not values:
            console.print(f"No key with the name {key} found")
        else:
            for path, value in values:
                console.print((f"Value: [red]{value}[/red], "
                               f"found in [red]/{'/'.join(path)}[/red]"))


@config.command()
@click.option("--key", "-k", required=True, help="the key (or parts thereof) which should be searched")
@click.option("--value", "-v", help="value the key should get associated with")
def set(key, value):
    """\b
    Sets a config key in the user configuration
    \b
    Loosly searches a config key and if multiple are found asks which key and value should be written.
    The user may pass a value to bypass prompting.
    \b
    Examples: 
    ovos-config set -k gui                              # lists all config keys containing "gui" (either as endpoint or in path),
                                                        # let the user choose the specific key and asks for the value
    ovos-config set -k blacklisted_skills -v myskill    # Adds "myskill" as an blacklisted skill
                                                        # Since this is a pretty specific key and a value is passed, the user won't be prompted
    """
    tuples = list(walkDict(CONFIG, key, full_path=True))
    values = [tup[1] for tup in tuples]
    paths = ["/".join(tup[0]) for tup in tuples]
    
    if len(paths) > 1:
        table = Table(show_header=True, header_style="bold red")
        table.add_column("#")
        table.add_column("Path", min_width=20)
        table.add_column("Value")

        for i, path in enumerate(paths):
            table.add_row(str(i), path, str(values[i]))
        console.print(table)

        exit_ = str(len(paths))
        choice = Prompt.ask(f"Which value should be changed? ({exit_}='Exit')",
                            choices=[str(i) for i in range(0, len(paths)+1)])
        if choice == exit_:
            console.print("User exit", style="red")
            exit()
    elif not paths:
        console.print(f"[red]Error:[/red] No key that fits the query")
        exit()
    else:
        choice = 0

    selected_path = paths[int(choice)]
    selected_value = values[int(choice)]
    selected_type = selected_value.__class__.__name__
    # to not irritate the use to suggest typing `["xyz"]`
    if selected_type == "list":
        selected_type = "str"
    
    if not value:
        value = Prompt.ask(("Please enter the value to be stored "
                            f"(type: [red]{selected_type}[/red]) "))
        value = value.replace('"','').replace("'","").replace("`","")

    # type checking/casting
    try:
        if isinstance(selected_value, str):
            _value = value
        elif isinstance(selected_value, bool):
            if value in ["true", "True", "1", "on"]:
                _value = True
            elif value in ["false", "False", "0", "off"]:
                _value = False
            else:
                raise TypeError
        elif isinstance(selected_value, list):
            try:
                _value = pathGet(local_conf, selected_path)
            except KeyError:
                _value = list()
                console.print(("Note: defining lists in the user config "
                               "will override subsequent list configurations"),
                               style="grey53")
            _value.append(value)
        elif isinstance(selected_value, int):
            _value = int(value)
        elif isinstance(selected_value, float):
            _value = float(value)
    except (TypeError, ValueError):
        console.print(f"[red]Error:[/red] The value passed can't be cast into {selected_type}")
        exit()
    
    local_conf = CONFIGS[2][1]
    pathSet(local_conf, selected_path, _value)
    local_conf.store()

if __name__ == "__main__":
    config()



# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
from os.path import isfile
from typing import Optional
from time import sleep

from ovos_config.models import LocalConf, MycroftDefaultConfig, \
    MycroftSystemConfig, MycroftUserConfig, RemoteConf
from ovos_config.locations import OLD_USER_CONFIG, get_xdg_config_save_path, \
    get_xdg_config_locations
from ovos_config.utils import FileWatcher

from ovos_utils.json_helper import flattened_delete, merge_dict
from ovos_utils.log import LOG


def _log_old_location_deprecation(old_user_config=OLD_USER_CONFIG):
    LOG.warning(" ===============================================")
    LOG.warning(" ==             DEPRECATION WARNING           ==")
    LOG.warning(" ===============================================")
    LOG.warning(f" You still have a config file at {old_user_config}")
    LOG.warning(" Note that this location is deprecated and will" +
                " not be used in the future")
    LOG.warning(" Please move it to " + get_xdg_config_save_path())


class Configuration(dict):
    """Namespace for operations on the configuration singleton."""
    __patch = LocalConf(None)  # Patch config that skills can update to override config
    bus = None
    default = MycroftDefaultConfig()
    system = MycroftSystemConfig()
    remote = RemoteConf()
    # This includes both the user config and
    # /etc/xdg/mycroft/mycroft.conf
    xdg_configs = [LocalConf(p) for p in get_xdg_config_locations()]
    _old_user = LocalConf(OLD_USER_CONFIG)
    # deprecation warning
    if isfile(OLD_USER_CONFIG):
        _log_old_location_deprecation(OLD_USER_CONFIG)
    _watchdog = None
    _callbacks = []

    def __init__(self):
        # python does not support proper overloading
        # when instantiation a Configuration object (new style)
        # the get method is replaced for proper dict behaviour
        self.get = self._real_get
        super().__init__(**self.load_all_configs())

    # dict methods
    def __setitem__(self, key, value):
        Configuration.__patch[key] = value
        super().__setitem__(key, value)
        # sync with other processes connected to bus
        if Configuration.bus:
            # imported from ovos_utils to allow FakeMessage if ovos-bus-client is missing
            from ovos_utils.messagebus import Message
            Configuration.bus.emit(Message("configuration.patch",
                                           {"config": {key: value}}))

    def __getitem__(self, item):
        super().update(Configuration.load_all_configs())
        return super().get(item)

    def __str__(self):
        super().update(Configuration.load_all_configs())
        try:
            return json.dumps(self, sort_keys=True)
        except:
            return super().__str__()

    def __dict__(self):
        super().update(Configuration.load_all_configs())
        return self

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        super().update(Configuration.load_all_configs())
        for k in super().__iter__():
            yield k

    def update(self, *args, **kwargs):
        Configuration.__patch.update(*args, **kwargs)
        super().update(*args, **kwargs)

    def pop(self, key):
        # we can not pop the key because configs are read only
        # we could do it for __patch but that does not make sense
        # for the object as a whole which is
        # supposed to behave like a python dict
        self.__setitem__(key, None)

    def items(self):
        super().update(Configuration.load_all_configs())
        return super().items()

    def keys(self):
        super().update(Configuration.load_all_configs())
        return super().keys()

    def values(self):
        super().update(Configuration.load_all_configs())
        return super().values()

    # config methods
    @staticmethod
    def load_config_stack(configs=None, cache=False, remote=True):
        """Load a stack of config dicts into a single dict

        Args:
            configs (list): list of dicts to load
            cache (boolean): True if result should be cached
            remote (boolean): False if the Mycroft Home settings shouldn't
                              be loaded
        Returns:
            (dict) merged dict of all configuration files
        """
        LOG.warning("load_config_stack has been deprecated, use load_all_configs instead")
        if configs:
            return Configuration.filter_and_merge(configs)
        system_constraints = Configuration.get_system_constraints()
        if not remote:
            system_constraints["disable_remote_config"] = True
        return Configuration.load_all_configs(system_constraints)

    @staticmethod
    def reset():
        """
        Remove any configuration patches and reload configuration
        """
        Configuration.__patch = {}
        Configuration.reload()

    @staticmethod
    def reload():
        """
        Reload all configuration files
        """
        Configuration.default.reload()
        Configuration.system.reload()
        Configuration.remote.reload()
        for cfg in Configuration.xdg_configs:
            cfg.reload()

    @staticmethod
    def get_system_constraints() -> dict:
        """
        Get Configuration constraints. Constraints must come from SYSTEM config.
        If not defined, then load the DEFAULT constraints.
        These settings can not be set anywhere else!
        @return: dict of system configuration constraints
        """

        return Configuration.system.get("system") or \
            Configuration.default.get("system") or \
            {}

    @staticmethod
    def load_all_configs(system_constraints: Optional[dict] = None) -> dict:
        """
        Load the stack of config files into a single dict
        @param system_constraints: constraints to limit user/remote config usage
        @return: merged dict of all configuration files
        """
        # system administrators can define different constraints in how
        # configurations are loaded
        system_constraints = system_constraints or \
            Configuration.get_system_constraints()
        skip_user = system_constraints.get("disable_user_config", False)
        skip_remote = system_constraints.get("disable_remote_config", False)

        configs = [Configuration.default, Configuration.system]
        if not skip_remote:
            configs.insert(1, Configuration.remote)
        if not skip_user:
            # deprecation warning
            if isfile(OLD_USER_CONFIG):
                configs.append(Configuration._old_user)
            configs += Configuration.xdg_configs

        # runtime patches by skills / bus events
        configs.append(Configuration.__patch)

        # Merge all configs into one
        return Configuration.filter_and_merge(configs)

    @staticmethod
    def filter_and_merge(configs) -> dict:
        """
        Build and return a configuration dict based on configuration files
        @param configs: List of Configuration objects to load
        @return: dict Configuration, built from `configs`
        """
        # ensure type
        for index, item in enumerate(configs):
            if isinstance(item, str):
                configs[index] = LocalConf(item)
            elif not isinstance(item, LocalConf):
                configs[index] = LocalConf(None)
                configs[index].merge(item)

        # system administrators can define different constraints in how
        # configurations are loaded
        system_conf = Configuration.get_system_constraints()
        protected_keys = system_conf.get("protected_keys") or {}
        protected_remote = protected_keys.get("remote") or []
        protected_user = protected_keys.get("user") or []
        skip_user = system_conf.get("disable_user_config", False)
        skip_remote = system_conf.get("disable_remote_config", False)

        # Merge all configs into one
        base = {}
        for cfg in configs:
            is_user = cfg.path is None or cfg.path not in [Configuration.default.path,
                                                           Configuration.system.path]
            is_remote = cfg.path == Configuration.remote.path
            if (is_remote and skip_remote) or (is_user and skip_user):
                continue
            elif is_remote:
                # delete protected keys from remote config
                for protection in protected_remote:
                    flattened_delete(cfg, protection)
            elif is_user:
                # delete protected keys from user config
                for protection in protected_user:
                    flattened_delete(cfg, protection)
            merge_dict(base, cfg)
        return base

    @staticmethod
    def set_config_update_handlers(bus):
        """
        Setup websocket handlers to update config on emitted changes.
        @param bus: Message bus client instance
        """
        from ovos_utils.network_utils import is_connected

        # remove any old event listeners
        Configuration.deregister_bus()

        # attach new bus and listeners
        Configuration.bus = bus
        bus.on("configuration.updated", Configuration.updated)
        bus.on("configuration.patch", Configuration.patch)
        bus.on("configuration.patch.clear", Configuration.patch_clear)
        bus.on("configuration.cache.clear", Configuration.clear_cache)
        # TODO unify these namespaces, they seem to differ between dev/mk2/PHAL
        bus.on("mycroft.paired", Configuration.handle_remote_update)
        bus.on("mycroft.internet.connected", Configuration.handle_remote_update)

        Configuration.set_config_watcher()

        # do the initial remote fetch
        if is_connected():
            Configuration.remote.reload()

    @staticmethod
    def set_config_watcher(callback: Optional[callable] = None):
        """
        Setup filewatcher to monitor for config file changes
        @param callback: optional method to call when configuration is changed
        """
        paths = [Configuration.system.path] + \
                [c.path for c in Configuration.xdg_configs]
        if callback and callback not in Configuration._callbacks:
            Configuration._callbacks.append(callback)
        if not Configuration._watchdog:
            Configuration._watchdog = FileWatcher(
                [p for p in paths if isfile(p)],
                Configuration._on_file_change
            )

    @staticmethod
    def _on_file_change(path: str):
        """
        Callback method for FileWatcher
        @param path: Configuration file path reporting a change
        """
        # reload updated config
        for cfg in Configuration.xdg_configs + [Configuration.system,
                                                Configuration.remote]:
            if cfg.path == path:
                old_cfg = hash(cfg)
                try:
                    cfg.reload()
                    reloaded = True
                except:
                    # got the file changed signal before write finished
                    sleep(0.5)
                    reloaded = False
                try:
                    if not reloaded:
                        LOG.warning(f"Reload failed, retrying")
                        cfg.reload()
                except:
                    LOG.exception("Failed to load configuration, "
                                  "syntax seems invalid!")
                new_cfg = hash(cfg)
                if old_cfg == new_cfg:
                    LOG.info(f"{path} unchanged")
                    return
                break
        else:
            LOG.info(f"Ignoring non-config file change: {path}")
            return

        LOG.info(f'{path} changed on disk')
        LOG.debug(f"Calling {len(Configuration._callbacks)} callbacks")
        for handler in Configuration._callbacks:
            try:
                handler()
            except:
                LOG.exception("Error in config update callback handler")

    @staticmethod
    def deregister_bus():
        """
        Remove messagebus handlers for configuration updates
        """
        if Configuration.bus:
            Configuration.bus.remove("configuration.updated",
                                     Configuration.updated)
            Configuration.bus.remove("configuration.patch",
                                     Configuration.patch)
            Configuration.bus.remove("configuration.patch.clear",
                                     Configuration.patch_clear)
            Configuration.bus.remove("configuration.cache.clear",
                                     Configuration.clear_cache)
            Configuration.bus.remove("mycroft.paired",
                                     Configuration.handle_remote_update)
            Configuration.bus.remove("mycroft.internet.connected",
                                     Configuration.handle_remote_update)

    @staticmethod
    def handle_remote_update(message):
        """Handler for paired/internet connect

        Triggers an update of remote config.
        """
        Configuration.remote.reload()

    @staticmethod
    def updated(message):
        """Handler for configuration.updated,

        Triggers an update of cached config.
        """
        Configuration.reload()

    @staticmethod
    def patch(message):
        """Patch the volatile dict usable by skills

        Args:
            message: Messagebus message should contain a config
                     in the data payload.
        """
        config = message.data.get("config", {})
        for k, v in config.items():
            Configuration.__patch[k] = v

    @staticmethod
    def patch_clear(message):
        """Clear the config patch space.

        Args:
            message: Messagebus message should contain a config
                     in the data payload.
        """
        Configuration.__patch = {}

    # Backwards compat methods
    @staticmethod
    def get(configs=None, cache=True, remote=True):
        """DEPRECATED - use Configuration class instead"""
        LOG.warning("Configuration.get() has been deprecated, use Configuration() instead")
        # NOTE: this is only called if using the class directly
        # if using an instance (dict object) self._real_get is called instead
        return Configuration.load_config_stack(configs, cache, remote)

    def _real_get(self, key, default=None):
        return self.__getitem__(key) or default

    @staticmethod
    def clear_cache(message=None):
        """DEPRECATED - there is no cache anymore """
        Configuration.updated(message)


def read_mycroft_config():
    """ returns a stateless dict with the loaded configuration """
    return dict(Configuration())


def update_mycroft_config(config, path=None, bus=None):
    """ updates user config file with the contents of provided dict
    if a path is provided that location will be used instead of MycroftUserConfig"""
    if path is None:
        conf = MycroftUserConfig()
    else:
        conf = LocalConf(path)
    conf.merge(config)
    conf.store()
    if bus:  # inform all Configuration objects connected to the bus
        # imported from ovos_utils to allow FakeMessage if ovos-bus-client is missing
        from ovos_utils.messagebus import Message
        bus.emit(Message("configuration.patch",  {"config": config}))
    return conf



"""This file will check a system level OpenVoiceOS specific config file

The ovos config is json with comment support like the regular mycroft.conf

Default locations tried by order until a file is found
- /etc/OpenVoiceOS/ovos.conf
- /etc/mycroft/ovos.conf

XDG locations are then merged over the select default config (if found)

Examples config:

{
   // the "name of the core",
   //         eg, OVOS, Neon, Chatterbox...
   //  all XDG paths should respect this
   //        {xdg_path}/{base_folder}/some_resource
   "base_folder": "OpenVoiceOS",

   // the filename of "mycroft.conf",
   //      eg, ovos.conf, chatterbox.conf, neon.conf...
   // "mycroft.conf" paths are derived from this
   //        {xdg_path}/{base_folder}/{config_filename}
   "config_filename": "mycroft.conf",

   // override the default.conf location, allows changing the default values
   //     eg, disable backend, disable skills, configure permissions
   "default_config_path": "/etc/OpenVoiceOS/default_mycroft.conf",

   // this is intended for derivative products, if ovos is being imported 
   // from one of these modules then the values below will be used instead
   //     eg, mycroft/ovos/neon can coexist in the same machine
   "module_overrides": {
        "neon_core": {
            "base_folder": "neon",
            "config_filename": "neon.conf",
            "default_config_path": "/opt/neon/neon.conf"
        }
   },
   // essentially aliases for the above, useful for microservice architectures
   "submodule_mappings": {
        "neon_speech": "neon_core",
        "neon_audio": "neon_core",
        "neon_enclosure": "neon_core"
   }
}
"""
import os
from os.path import isfile, join, dirname

from json_database import JsonStorage

import ovos_config.locations as _oloc
from ovos_utils.json_helper import load_commented_json, merge_dict
from ovos_utils.log import LOG


def get_ovos_config():
    """
    Goes through all possible ovos.conf paths and loads them in order. Default
    `base_folder` and `config_filename` are overridden by envvars
    `OVOS_CONFIG_BASE_FOLDER` and `OVOS_CONFIG_FILENAME`, respectively.
    Submodule overrides are applied to the final config if defined for the
    calling module.
    eg, if neon is calling this method then neon config overrides are loaded

    """
    from ovos_utils.system import is_running_from_module

    # populate default values
    config = {"xdg": True,
              "base_folder": os.environ.get("OVOS_CONFIG_BASE_FOLDER") or
              "mycroft",
              "config_filename": os.environ.get("OVOS_CONFIG_FILENAME") or
              "mycroft.conf"}
    try:
        config["default_config_path"] = _oloc.find_default_config()
    except FileNotFoundError:  # not a mycroft device
        config["default_config_path"] = join(dirname(__file__), "mycroft.conf")

    # load ovos.conf
    for path in get_ovos_default_config_paths():
        try:
            config = merge_dict(config, load_commented_json(path))
        except:
            # tolerate bad json TODO proper exception (?)
            pass

    # let's check for derivatives specific configs
    # the assumption is that these cores are exclusive to each other,
    # this will never find more than one override
    # TODO this works if using dedicated .venvs what about system installs?
    cores = config.get("module_overrides") or {}
    for k in cores:
        if is_running_from_module(k):
            config = merge_dict(config, cores[k])
            break
    else:
        subcores = config.get("submodule_mappings") or {}
        for k in subcores:
            if is_running_from_module(k):
                config = merge_dict(config, cores[subcores[k]])
                break

    return config


def save_ovos_config(new_config):
    """ update ovos.conf contents at ~/.config/OpenVoiceOS/ovos.conf """
    OVOS_CONFIG = join(_oloc.get_xdg_config_save_path("OpenVoiceOS"),
                       "ovos.conf")
    cfg = JsonStorage(OVOS_CONFIG)
    cfg.update(new_config)
    cfg.store()
    return cfg


def get_ovos_default_config_paths():
    """ return a list of all existing ovos.conf file locations by order of precedence

     eg. ["/etc/OpenVoiceOS/ovos.conf", "/home/user/.config/OpenVoiceOS/ovos.conf"]

     """
    paths = []
    if isfile("/etc/OpenVoiceOS/ovos.conf"):
        paths.append("/etc/OpenVoiceOS/ovos.conf")
    elif isfile("/etc/mycroft/ovos.conf"):
        LOG.warning("found /etc/mycroft/ovos.conf\n"
                    "This location has been DEPRECATED!\n"
                    "Please move your config to /etc/OpenVoiceOS/ovos.conf")
        paths.append("/etc/mycroft/ovos.conf")

    # This includes both the user config and
    # /etc/xdg/OpenVoiceOS/ovos.conf
    for p in _oloc.get_xdg_config_dirs("OpenVoiceOS"):
        if isfile(join(p, "ovos.conf")):
            paths.append(join(p, "ovos.conf"))

    return paths


def is_using_xdg():
    """ BACKWARDS COMPAT: logs warning and always returns True"""
    LOG.warning("is_using_xdg has been deprecated! XDG specs are always honoured, this method will be removed in a future release")
    return True


def get_xdg_base():
    """ base folder name to be used when building paths of the format {$XDG_XXX}/{base}

    different derivative cores may change this folder, this value is derived from ovos.conf
        eg, "mycroft", "hivemind", "neon" ....
    """

    return get_ovos_config().get("base_folder") or "mycroft"


def set_xdg_base(folder_name):
    """ base folder name to be used when building paths of the format {$XDG_XXX}/{base}

    different derivative cores may change this folder, this value is derived from ovos.conf
        eg, "mycroft", "hivemind", "neon" ....

    NOTE: this value will be set globally, per core overrides in ovos.conf take precedence
    """
    LOG.info(f"XDG base folder set to: '{folder_name}'")
    save_ovos_config({"base_folder": folder_name})


def set_config_filename(file_name, core_folder=None):
    """ base config file name to be used when building paths

    different derivative cores may change this filename, this value is derived from ovos.conf
        eg, "mycroft.conf", "hivemind.json", "neon.yaml" ....

    NOTE: this value will be set globally, per core overrides in ovos.conf take precedence
    """
    if core_folder:
        set_xdg_base(core_folder)
    LOG.info(f"config filename set to: '{file_name}'")
    save_ovos_config({"config_filename": file_name})


def get_config_filename():
    """ base config file name to be used when building paths

    different derivative cores may change this filename, this value is derived from ovos.conf
        eg, "mycroft.conf", "hivemind.json", "neon.yaml" ....
    """
    return get_ovos_config().get("config_filename") or "mycroft.conf"


def set_default_config(file_path=None):
    """ full path to default config file to be used
    NOTE: this is a full path, not a directory! "config_filename" parameter is not used here

    different derivative cores may change this file, this value is derived from ovos.conf

    NOTE: this value will be set globally, per core overrides in ovos.conf take precedence
    """
    file_path = file_path or _oloc.find_default_config()
    LOG.info(f"default config file changed to: {file_path}")
    save_ovos_config({"default_config_path": file_path})
